@@ -129,6 +129,10 @@ static struct algo ecc = {(CK_BYTE *)"RTCMK   ECC     ", (CK_BYTE *)"ECC", 2 };
 static struct algo rsa = {(CK_BYTE *)"RTCMK   ", (CK_BYTE *)"RSA", 1 };
 static struct algo ibm_dilithium = {(CK_BYTE *)"RTCMK   QSA     ",
                                                 (CK_BYTE *)"IBM Dilithium", 2 };
+static struct algo ibm_ml_dsa = {(CK_BYTE *)"RTCMK   QSA     ",
+                                                (CK_BYTE *)"IBM ML_DSA", 2 };
+static struct algo ibm_ml_kem = {(CK_BYTE *)"RTCMK   QSA     ",
+                                                (CK_BYTE *)"IBM ML-KEM", 2 };
 
 int cca_decrypt(unsigned char *in_data, unsigned long in_data_len,
                 unsigned char *out_data, unsigned long *out_data_len,
@@ -669,21 +673,27 @@ CK_FUNCTION_LIST *p11_init(void)
     if (getfunclist == NULL) {
         print_error("Couldn't get the address of the C_GetFunctionList "
                     "routine.");
+#ifndef WITH_SANITIZER
         dlclose(p11_lib);
+#endif
         return NULL;
     }
 
     rv = getfunclist(&funcs);
     if (rv != CKR_OK) {
         p11_error("C_GetFunctionList", rv);
+#ifndef WITH_SANITIZER
         dlclose(p11_lib);
+#endif
         return NULL;
     }
 
     rv = funcs->C_Initialize(NULL_PTR);
     if (rv != CKR_OK) {
         p11_error("C_Initialize", rv);
+#ifndef WITH_SANITIZER
         dlclose(p11_lib);
+#endif
         return NULL;
     }
 
@@ -697,8 +707,10 @@ void p11_fini(CK_FUNCTION_LIST *funcs)
 {
     funcs->C_Finalize(NULL_PTR);
 
+#ifndef WITH_SANITIZER
     if (p11_lib)
         dlclose(p11_lib);
+#endif
 }
 
 /* Expect attribute array to have 3 entries,
@@ -729,6 +741,8 @@ int add_key(CK_OBJECT_HANDLE handle, CK_ATTRIBUTE *attrs, struct key **keys)
     case CKK_GENERIC_SECRET:
     case CKK_RSA:
     case CKK_IBM_PQC_DILITHIUM:
+    case CKK_IBM_ML_DSA:
+    case CKK_IBM_ML_KEM:
         break;
     default:
         free(new_key);
@@ -749,6 +763,7 @@ int add_key(CK_OBJECT_HANDLE handle, CK_ATTRIBUTE *attrs, struct key **keys)
     new_key->label = malloc(attrs[2].ulValueLen + 1);
     if (!new_key->label) {
         print_error("Malloc of %lu bytes failed!", attrs[2].ulValueLen + 1);
+        free(new_key->opaque_attr);
         free(new_key);
         return 2;
     }
@@ -789,6 +804,12 @@ int add_key(CK_OBJECT_HANDLE handle, CK_ATTRIBUTE *attrs, struct key **keys)
         case CKK_IBM_PQC_DILITHIUM:
             type_name = IBM_DILITHIUM_NAME;
             break;
+        case CKK_IBM_ML_DSA:
+            type_name = IBM_ML_DSA_NAME;
+            break;
+        case CKK_IBM_ML_KEM:
+            type_name = IBM_ML_KEM_NAME;
+            break;
         default:
             type_name = BAD_NAME;
         }
@@ -1024,6 +1045,7 @@ int cca_migrate_symmetric(struct key *key, char **out, struct algo algo)
 {
     long return_code, reason_code, exit_data_length;
     unsigned char *key_identifier;
+    long key_identifier_length;
 
     exit_data_length = 0;
 
@@ -1034,24 +1056,49 @@ int cca_migrate_symmetric(struct key *key, char **out, struct algo algo)
     }
     memcpy(key_identifier, (char *) key->opaque_attr, key->attr_len);
 
-    CSNBKTC(&return_code,
-            &reason_code,
-            &exit_data_length,
-            NULL, &(algo.rule_array_count), algo.rule_array, key_identifier);
+    if (key_identifier[0] == 0x01 && key_identifier[4] == 0x05 &&
+        key_identifier[41] == 0x02) {
+        /* AES CIPHER key */
+        key_identifier_length = key->type == CKK_AES_XTS ?
+                                        key->attr_len / 2 : key->attr_len;
+        CSNBKTC2(&return_code,
+                 &reason_code,
+                 &exit_data_length,
+                 NULL,
+                 &(algo.rule_array_count),
+                 algo.rule_array, &key_identifier_length, key_identifier);
 
-    if (return_code != CCA_SUCCESS) {
-        cca_error("CSNBKTC (Key Token Change)", return_code, reason_code);
-        print_error("Migrating %s key failed. label=%s, handle=%lu",
-                    algo.name, key->label, key->handle);
-        return 1;
-    }
+        if (return_code != CCA_SUCCESS) {
+            cca_error("CSNBKTC2 (Key Token Change)", return_code, reason_code);
+            print_error("Migrating %s key failed. label=%s, handle=%lu",
+                        algo.name, key->label, key->handle);
+            return 1;
+        }
 
-    if (key->type == CKK_AES_XTS) {
+        if (key->type == CKK_AES_XTS) {
+            key_identifier_length = key->attr_len / 2;
+            CSNBKTC2(&return_code,
+                     &reason_code,
+                     &exit_data_length,
+                     NULL,
+                     &(algo.rule_array_count),
+                     algo.rule_array, &key_identifier_length,
+                     key_identifier + key_identifier_length);
+
+            if (return_code != CCA_SUCCESS) {
+                cca_error("CSNBKTC2 (Key Token Change)", return_code,
+                          reason_code);
+                print_error("Migrating %s key failed. label=%s, handle=%lu",
+                            algo.name, key->label, key->handle);
+                return 1;
+            }
+        }
+    } else {
+        /* AES DATA key */
         CSNBKTC(&return_code,
                 &reason_code,
                 &exit_data_length,
-                NULL, &(algo.rule_array_count), algo.rule_array,
-                key_identifier + (key->attr_len / 2));
+                NULL, &(algo.rule_array_count), algo.rule_array, key_identifier);
 
         if (return_code != CCA_SUCCESS) {
             cca_error("CSNBKTC (Key Token Change)", return_code, reason_code);
@@ -1059,8 +1106,23 @@ int cca_migrate_symmetric(struct key *key, char **out, struct algo algo)
                         algo.name, key->label, key->handle);
             return 1;
         }
-    }
 
+        if (key->type == CKK_AES_XTS) {
+            CSNBKTC(&return_code,
+                    &reason_code,
+                    &exit_data_length,
+                    NULL, &(algo.rule_array_count), algo.rule_array,
+                    key_identifier + (key->attr_len / 2));
+
+            if (return_code != CCA_SUCCESS) {
+                cca_error("CSNBKTC (Key Token Change)", return_code,
+                          reason_code);
+                print_error("Migrating %s key failed. label=%s, handle=%lu",
+                            algo.name, key->label, key->handle);
+                return 1;
+            }
+        }
+    }
     if (v_level) {
         printf("Successfully migrated %s key. label=%s, handle=%lu\n",
                algo.name, key->label, key->handle);
@@ -1183,6 +1245,22 @@ int cca_migrate(struct key *keys, struct key_count *count,
             else
                 count->ibm_dilithium++;
             break;
+        case CKK_IBM_ML_DSA:
+            rc = cca_migrate_asymmetric(key, &migrated_data, ibm_ml_dsa,
+                                        masterkey);
+            if (rc)
+                count_failed->ibm_ml_dsa++;
+            else
+                count->ibm_ml_dsa++;
+            break;
+        case CKK_IBM_ML_KEM:
+            rc = cca_migrate_asymmetric(key, &migrated_data, ibm_ml_kem,
+                                        masterkey);
+            if (rc)
+                count_failed->ibm_ml_kem++;
+            else
+                count->ibm_ml_kem++;
+            break;
         default:
             rc = 1;
             break;
@@ -1233,7 +1311,8 @@ done:
 void key_migration_results(struct key_count migrated, struct key_count failed)
 {
     if (migrated.aes || migrated.des || migrated.des2 || migrated.des3 ||
-        migrated.ecc || migrated.hmac || migrated.rsa || migrated.ibm_dilithium)
+        migrated.ecc || migrated.hmac || migrated.rsa ||
+        migrated.ibm_dilithium || migrated.ibm_ml_dsa || migrated.ibm_ml_kem)
         printf("Successfully migrated: ");
     if (migrated.aes)
         printf("AES: %d. ", migrated.aes);
@@ -1251,9 +1330,14 @@ void key_migration_results(struct key_count migrated, struct key_count failed)
         printf("RSA: %d. ", migrated.rsa);
     if (migrated.ibm_dilithium)
         printf("IBM Dilithium: %d. ", migrated.ibm_dilithium);
+    if (migrated.ibm_ml_dsa)
+        printf("IBM ML-DSA: %d. ", migrated.ibm_ml_dsa);
+    if (migrated.ibm_ml_kem)
+        printf("IBM ML-KEM: %d. ", migrated.ibm_ml_kem);
 
     if (failed.aes || failed.des || failed.des2 || failed.des3 ||
-        failed.ecc || failed.hmac || failed.rsa || failed.ibm_dilithium)
+        failed.ecc || failed.hmac || failed.rsa ||
+        failed.ibm_dilithium || failed.ibm_ml_dsa || failed.ibm_ml_kem)
         printf("\nFailed to migrate: ");
     if (failed.aes)
         printf("AES: %d. ", failed.aes);
@@ -1271,6 +1355,10 @@ void key_migration_results(struct key_count migrated, struct key_count failed)
         printf("RSA: %d. ", failed.rsa);
     if (failed.ibm_dilithium)
         printf("IBM Dilithium: %d. ", failed.ibm_dilithium);
+    if (failed.ibm_ml_dsa)
+        printf("IBM ML-DSA: %d. ", failed.ibm_ml_dsa);
+    if (failed.ibm_ml_kem)
+        printf("IBM ML-KEM: %d. ", failed.ibm_ml_kem);
 
     printf("\n");
 }
@@ -1281,8 +1369,8 @@ int migrate_wrapped_keys(CK_SLOT_ID slot_id, const char *userpin, int masterkey)
     CK_KEY_TYPE key_type = 0;
     CK_SESSION_HANDLE sess;
     CK_RV rv;
-    struct key_count count = { 0, 0, 0, 0, 0, 0, 0, 0 };
-    struct key_count count_failed = { 0, 0, 0, 0, 0, 0, 0, 0 };
+    struct key_count count = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
+    struct key_count count_failed = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
     int exit_code = 0, rc;
 
     funcs = p11_init();
@@ -1355,6 +1443,22 @@ int migrate_wrapped_keys(CK_SLOT_ID slot_id, const char *userpin, int masterkey)
         if (rc) {
             goto done;
         }
+        if (v_level)
+            printf("Search for IBM ML-DSA keys\n");
+        key_type = CKK_IBM_ML_DSA;
+        rc = migrate_keytype(funcs, sess, &key_type, &count, &count_failed,
+                             masterkey);
+        if (rc) {
+            goto done;
+        }
+        if (v_level)
+            printf("Search for IBM ML-KEM keys\n");
+        key_type = CKK_IBM_ML_KEM;
+        rc = migrate_keytype(funcs, sess, &key_type, &count, &count_failed,
+                             masterkey);
+        if (rc) {
+            goto done;
+        }
         break;
     case MK_ASYM:
         if (v_level)
@@ -1699,7 +1803,7 @@ static int verbose_str2level(char *str)
 
 int main(int argc, char **argv)
 {
-    int ret = -1, opt = 0, c_flag = 0, masterkey = 0;
+    int ret = -1, opt = 0, masterkey = 0;
     int data_store_len = 0;
     CK_SLOT_ID slot_id = 0;
     const char *sopin = NULL, *userpin = NULL;
@@ -1762,7 +1866,6 @@ int main(int argc, char **argv)
             }
             break;
         case 's':
-            c_flag++;
             slot_id = atoi(optarg);
             break;
         case 'v':
